/*
 *  linux/fs/ods2/ods2.h
 *
 * COPYRIGHT
 *      This file is distributed under the terms of the GNU General Public
 *      License (GPL). Copies of the GPL can be obtained from:
 *              ftp://prep.ai.mit.edu/pub/gnu/GPL
 *      Each contributing author retains all rights to their own work.
 *
 * Written 2003 by Jonas Lindholm <jlhm@usa.net>
 *
 */

/*
  The followinf structures are defined in the book
  "VMS File System Internals"
*/

/*
  This is the home block on a ODS2 disk.
*/

typedef struct hm2def {
	long unsigned       hm2_l_homelbn;
	long unsigned       hm2_l_alhomelbn;
	long unsigned       hm2_l_altidxlbn;
	union {
		short unsigned    hm2_w_struclev;
		struct {
			char unsigned   hm2_b_structlevv;
			char unsigned   hm2_b_structlevl;
		} s1;
	} u1;
	short unsigned      hm2_w_cluster;
	short unsigned      hm2_w_homevbn;
	short unsigned      hm2_w_alhomevbn;
	short unsigned      hm2_w_altidxvbn;
	short unsigned      hm2_w_ibmapvbn;
	long unsigned       hm2_l_ibmaplbn;
	long unsigned       hm2_l_maxfiles;
	short unsigned      hm2_w_ibmapsize;
	short unsigned      hm2_w_resfiles;
	short unsigned      hm2_w_devtype;
	short unsigned      hm2_w_rvn;
	short unsigned      hm2_w_setcount;
	short unsigned      hm2_w_volchar;
	union {
		long unsigned     hm2_l_owner;
		struct {
			short unsigned  hm2_w_mem;
			short unsigned  hm2_w_grp;
		} s1;
	} u2;
	long unsigned       hm2_l_res1;
	short unsigned      hm2_w_protect;
	short unsigned      hm2_w_fileprot;
	short unsigned      hm2_w_res2;
	short unsigned      hm2_w_checksum1;
	long unsigned       hm2_q_credate[2];
	char unsigned       hm2_b_window;
	char unsigned       hm2_b_lru_lim;
	short unsigned      hm2_w_extend;
	long unsigned       hm2_q_retainmin[2];
	long unsigned       hm2_q_retainmax[2];
	long unsigned       hm2_q_revdate[2];
	char unsigned       hm2_r_min_class[20];
	char unsigned       hm2_r_max_class[20];
	char unsigned       hm2_b_res3[320];
	long unsigned       hm2_l_serialnum;
	char                hm2_t_structname[12];
	char                hm2_t_volname[12];
	char                hm2_t_ownername[12];
	char                hm2_t_format[12];
	short unsigned      hm2_w_res4;
	short unsigned      hm2_w_checksum2;
} HM2DEF;

/*
  This is the Storage Control Block.
  It is the first block in file BITMAP.SYS.
*/

typedef struct scbdef {
	union {
		short unsigned    scb_w_struclev;
		struct {
			char unsigned   scb_b_structlevv;
			char unsigned   scb_b_structlevl;
		} s1;
	} u1;
	short unsigned      scb_w_cluster;
	long unsigned       scb_l_volsize;
	long unsigned       scb_l_blksize;
	long unsigned       scb_l_sectors;
	long unsigned       scb_l_tracks;
	long unsigned       scb_l_cylinders;
	union {
		long unsigned     scb_l_status;
		struct {
			long unsigned   scb_v_mapdirty:1;
			long unsigned   scb_v_mapalloc:1;
			long unsigned   scb_v_filalloc:1;
			long unsigned   scb_v_quodirty:1;
			long unsigned   scb_v_hdrwrite:1;
			long unsigned   scb_v_corrupt:1;
		} s1;
	} u2;
	union {
		long unsigned     scb_l_status2;
		struct {
			long unsigned   scb_v_mapdirty:1;
			long unsigned   scb_v_mapalloc:1;
			long unsigned   scb_v_filalloc:1;
			long unsigned   scb_v_quodirty:1;
			long unsigned   scb_v_hdrwrite:1;
			long unsigned   scb_v_corrupt:1;
		} s1;
	} u3;
	short unsigned      scb_w_writecnt;
	char                scb_t_volockname[12];
	short unsigned      scb_q_mounttime[4];
	short unsigned      scb_w_backrev;
	long long unsigned  scb_q_genernum;
	char unsigned       scb_b_reserved[446];
	short unsigned      scb_w_checksum;
} SCBDEF;

/*
  This structure is part of the file header block and
  fives different tomes as well as the file name.
*/

typedef struct fi2def {
	char                fi2_t_filename[20];
	short unsigned      fi2_w_revision;
	short unsigned      fi2_q_credate[4];
	short unsigned      fi2_q_revdate[4];
	short unsigned      fi2_q_expdate[4];
	short unsigned      fi2_q_bakdate[4];
	char                fi_2_filenameext[66];
} FI2DEF;

/*
  This is the file header for any ODS2 file.
  It is located in file INDEXF.SYS.
*/

typedef struct fh2def {
	char unsigned       fh2_b_idoffset;
	char unsigned       fh2_b_mpoffset;
	char unsigned       fh2_b_acoffset;
	char unsigned       fh2_b_rsoffset;
	short unsigned      fh2_w_seg_num;
	union {
		short unsigned    fh2_w_struclev;
		struct {
			char unsigned   fh2_b_structlevv;
			char unsigned   fh2_b_structlevl;
		} s1;
	} u1;
	union {
		short unsigned    fh2_w_fid[3];
		struct {
			short unsigned  fh2_w_fid_num;
			short unsigned  fh2_w_fid_seq;
			char unsigned   fh2_b_fid_rvn;
			char unsigned   fh2_b_fid_nmx;
		} s1;
	} u2;
	union {
		short unsigned    fh2_w_ext_fid[3];
		struct {
			short unsigned  fid_w_ex_fidnum;
			short unsigned  fid_w_ex_fidseq;
			char unsigned   fid_b_ex_fidrvn;
			char unsigned   fid_b_ex_fidnmx;
		} s1;
	} u3;
	long unsigned       fh2_w_recattr[8];
	union {
		struct filechar {
			long unsigned   fch_v_wascontig:1;
			long unsigned   fch_v_nobackup:1;
			long unsigned   fch_v_writeback:1;
			long unsigned   fch_v_readcheck:1;
			long unsigned   fch_v_writecheck:1;
			long unsigned   fch_v_contigb:1;
			long unsigned   fch_v_locked:1;
			long unsigned   fch_v_contig:1;
			long unsigned   fch_v_res1:3;
			long unsigned   fch_v_badacl:1;
			long unsigned   fch_v_spool:1;
			long unsigned   fch_v_directory:1;
			long unsigned   fch_v_badblock:1;
			long unsigned   fch_v_markdel:1;
			long unsigned   fch_v_nocharge:1;
			long unsigned   fch_v_erase:1;
		} s1;
		long unsigned     fh2_l_filechar;
	} u4;
	short unsigned      fh2_w_res1;
	char unsigned       fh2_b_map_inuse;
	char unsigned       fh2_b_acc_mode;
	union {
		long unsigned     fh2_l_fileowner;
		struct {
			short unsigned  fh2_w_mem;
			short unsigned  fh2_w_grp;
		} s1;
	} u5;
	short unsigned      fh2_w_fileprot;
	union {
		short unsigned    fh2_w_backlink[3];
		struct {
			short unsigned  fid_w_num;
			short unsigned  fid_w_seq;
			char unsigned   fid_b_rvn;
			char unsigned   fid_b_nmx;
		} s1;
	} u6;
	char unsigned       fh2_b_journal;
	char unsigned       fh2_b_ru_active;
	short unsigned      fh2_w_res2;
	long unsigned       fh2_l_highwater;
	char unsigned       fh2_b_res3[8];
	char unsigned       fh2_r_class_prot[20];
	char unsigned       fh2_b_res4[402];
	short unsigned      fh2_w_checksum;
} FH2DEF;

/*
  This is the file attribute structure.
  It is part of the file header.
  It defines RMS attributes for any file.
*/

#define FAT_C_UNDEFINED  0
#define FAT_C_FIXED      1
#define FAT_C_VARIABLE   2
#define FAT_C_VFC        3
#define FAT_C_STREAM     4
#define FAT_C_STREAMLF   5
#define FAT_C_STREAMCR   6

#define FAT_C_SEQUANTIAL 0
#define FAT_C_RELATIVE   1
#define FAT_C_INDEXED    2
#define FAT_C_DIRECT     3

#define FAT_M_FORTRANCC  0x01
#define FAT_M_IMPLIEDCC  0x02
#define FAT_M_PRINTCC    0x04
#define FAT_M_NOSPAN     0x08
#define FAT_M_MSBRCW     0x10

typedef struct fatdef {
	union {
		char unsigned     fat_b_rtype;
		struct {
			char unsigned   fat_v_rtype:4;
			char unsigned   fat_v_fileorg:4;
		} s0;
	} u0;
	char unsigned       fat_b_rattrib;
	char unsigned       fat_w_rsize;
	union {
		long unsigned     fat_l_hiblk;
		struct {
			short unsigned  fat_w_hiblkh;
			short unsigned  fat_w_hiblkl;
		} s1;
	} u1;
	union {
		long unsigned     fat_l_efblk;
		struct {
			short unsigned  fat_w_efblkh;
			short unsigned  fat_w_efblkl;
		} s1;
	} u2;
	short unsigned      fat_w_ffbyte;
	char unsigned       fat_b_bktsize;
	char unsigned       fat_b_vfcsize;
	short unsigned      fat_w_maxrec;
	short unsigned      fat_w_defext;
	char unsigned       fat_b_res1[6];
	short unsigned      fat_w_notused;
	short unsigned      fat_w_versions;
} FATDEF;

/*
  This is the structure used for mapping virtual block
  number, VBN, to logical block numbers, LBN.
  One or more of this structure is part of the file header.
*/

typedef struct fm2def {
	union {
		struct {
			char unsigned  fm2_b_count1;
			char unsigned  fm2_v_highlbn:6;
			char unsigned  fm2_v_format:2;
			short unsigned fm2_w_lowlbn;
		} fm1;
		struct {
			short unsigned fm2_v_count2:14;
			short unsigned fm2_v_format:2;
			short unsigned  fm2_l_lbn2[2];
		} fm2;
		struct {
			short unsigned fm2_v_count2:14;
			short unsigned fm2_v_format:2;
			short unsigned fm2_w_lowcount;
			long unsigned  fm2_l_lbn3;
		} fm3;
	} u1;
} FM2DEF;

/*
  This structure define a directory entry in a directory file.
*/

#define DIR_C_FID           0
#define DIR_C_LINKNAME      1

typedef struct dirdef {
	union {
		struct {
			short unsigned     dir_w_size;
			short              dir_w_verlimit;
			union {
				char unsigned      dir_b_flags;
				struct {
					char unsigned dir_v_type:3;
					char unsigned dir_v_res1:3;
					char unsigned dir_v_nextrec:1;
					char unsigned dir_v_prevrec:1;
				} s4;
			} u4;
			char unsigned      dir_b_namecount;
			char               dir_t_name;
		} s1;
		struct {
			short unsigned     dir_w_version;
			union {
				short unsigned   dir_w_fid[3];
				struct {
					short unsigned fid_w_num;
					short unsigned fid_w_seq;
					char unsigned  fid_b_rvn;
					char unsigned  fid_b_nmx;
				} s3;
			} u2;
		} s2;
	} u1;
} DIRDEF;


/*
  From here we have our own ODS2 specific structures
  and definitions.
*/

typedef struct ods2map {
	struct ods2map     *nxt;
	struct {
		unsigned long     cnt;
		unsigned long     lbn;
	} s1[16];
} ODS2MAP;

/*
  Each block map 64Kbyte * 16 loff's.
  The number of bytes for this structure is 4 + 16 * 16 => 260.
  For a 1GB file we need a total of 1024 blocks. If each block is
  260 bytes the total amount of bytes is 1024 * 260 => 266240 bytes
  The linked list will contain no more than 8 blocks as the structure
  below has 128 pointers.
*/

typedef struct ods2var {
	struct ods2var     *nxt;     /* next block if needed */
	struct {
		loff_t            recoffs; /* offset to start of record */
		loff_t            loff;    /* virtual offset to start of record */
	} s1[16];
} ODS2VAR;

/*
  Each file that is of variable record type has the following structure
  attached to it.
  This is the index for one or more ODS2VAR structures. By doing index as
  much as possible it is easy to calculate what structure to use by just
  doing some shifts and bit masking.
  Note that this structure and its sub structures are protected by a 
  semaphore because more than one process at the same time can use the inode
  structure to read the file contents.
  The number of bytes for this structure is 128 * 4 + 12 (or 16) => 528.
  The overhead for small files are big but 528 bytes allocated using kmalloc
  should not be to much.
*/

#define IDXVAR(a)     ((a >> 16) & 0x0f)
#define IDXVARI(a)    ((a >> 20) & 0x7f)
#define IDXBLOCK(a)   (a >> 27)

typedef struct ods2vari {
	ODS2VAR            *ods2varp[128]; /* pointers to ods2var blocks */
	struct semaphore    sem;           /* This is the semaphore used to protect this structure */
	loff_t              highidx;       /* highest index so far... */
} ODS2VARI;

/*
  Each open file has the following structure attached to it.
  It add the extra variables needed to handle directories and
  RMS data.
*/

typedef struct ods2file {
	struct buffer_head *bhp;
	long long unsigned  currec; /* byte offset to current record --- from start of file */
	short unsigned      curbyte; /* byte offset into current record */
	short unsigned      reclen; /* length of current record */
	union {
		long unsigned       flags;
		struct {
			long unsigned     v_raw:1; /* this file handler must return data in raw mode */
			long unsigned     v_res1:31;
		} s1;
	} u1;
} ODS2FILE;

/*
  Each inode has the following structure attached to it.
  It keep the file attributes and mapping information in memory.
*/

typedef struct ods2fh {
	ODS2MAP            *map;      /* mapping information from VBN to LBN */
	ODS2VARI           *ods2vari; /* only used for variable record files */
	FATDEF              fat;      /* file attributes */
	long unsigned       parent;   /* ino of parent directory */
} ODS2FH;

/*
  The super block for an ODS2 disk has the following
  structure attached.
  It keep the home block and the inode for INDEXF.SYS;1
  in memory.
*/

#define SB_M_VERSALL       0
#define SB_M_VERSHIGH      1
#define SB_M_VERSNONE      2
#define SB_M_RAW           8
#define SB_M_LOWERCASE     16

typedef struct ods2sb {
	HM2DEF              hm2;
	struct inode       *indexf;  /* INDEXF.SYS */
	char unsigned      *ibitmap; /* index file header bitmap */
	struct statfs       statfs;
	struct {
		int         v_version:3; /* what to do with file versions */
		int         v_raw:1;     /* force all files as stream */
		int         v_lowercase:1; /* force all file names to lowercase */
		int         v_res:27;    /* reserved */
	} flags;
	char                dollar;      /* character used for dollar */
	char                semicolon;   /* character used for semicolon */
} ODS2SB;

/*
  This is our private ioctl operations for a file pointer.
*/

#define ODS2_IOC_FISETRAW  _IOW('f', 0x0d0, long) /* enable/disable raw file mode */
#define ODS2_IOC_FIGETRAW  _IOR('g', 0x0d0, long) /* get raw file mode */
#define ODS2_IOC_SBGETRAW  _IOR('g', 0x0d1, long) /* get raw mode for super block */

#define MIN(a, b) (a < b ? a : b)
#define MAX(a, b) (a > b ? a : b)

/*
  Ok, I give up :-) for some reason unknown to me the addition of 2 seconds
  is needed to get the correct time.
  It works for a file created 1-jan-1971 and for a file created 1-jan-2038
  as well as for files created 1992 and 2003.
*/

#define vms2unixtime(a) ((unsigned long)div64((le64_to_cpu(*(long long unsigned *)&(a)) - 0x007c953d63a19980L) >> 7, 78125) + 2);

/*
  util.c
*/

unsigned long long div64(unsigned long long a, unsigned long b0);
unsigned long vbn2lbn(struct super_block *sb, ODS2MAP *map, unsigned long vbn);
unsigned long ino2fhlbn(struct super_block *sb, unsigned long ino);
ODS2MAP *getmap(struct super_block *sb, FH2DEF *fh2p);
struct buffer_head *getfilebh(struct file *filp, long unsigned vbn);
int verify_fh(FH2DEF *fh2p, long unsigned ino);
int parse_options(struct super_block *sb, char *options);

/*
  inode.c
*/

struct dentry *ods2_lookup(struct inode *dir, struct dentry *dentry);
void ods2_read_inode(struct inode *inode);
void ods2_put_inode(struct inode *inode);
void ods2_clear_inode(struct inode *inode);
void ods2_delete_inode(struct inode *inode);

/*
  dir.c
*/

int ods2_readdir(struct file *filp, void *dirent, filldir_t filldir);

/*
  file.c
*/

int ods2_file_ioctl(struct inode *inode, struct file *filp, int unsigned cmd, long unsigned arg);
ssize_t ods2_read(struct file *filp, char *buf, size_t buflen, loff_t *loff);
loff_t ods2_llseek(struct file *filp, loff_t loff, int seek);
int ods2_open_release(struct inode *inode, struct file *filp);


/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
